@@ -1,7 +1,7 @@
 
 from .classifiers.linear import MulticlassLogisticRegression
 
-from .encoders.sklearnencoders import ExperimentalPCA, ExperimentalUMAP
+from .encoders.sklearnencoders import ExperimentalPCA, ExperimentalIncrementalPCA, ExperimentalUMAP
 from .encoders.dictembedding import DictEmbedding
 
 
@@ -12,6 +12,7 @@ classifiers_dict = {
 
 encoders_dict = {
     'PCA': ExperimentalPCA,
+    'IncrementalPCA': ExperimentalIncrementalPCA,
     'UMAP': ExperimentalUMAP,
     'embedding_dict': DictEmbedding
 }
